/*

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the Free
 Software Foundation; either version 2 of the License, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 more details.

 You should have received a copy of the GNU General Public License along with
 this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

*/

//
//   RS232C device file emulation module
//
//
//	Copyright(C) 2010 Brother Industries, Ltd.
//           All Rights Reserved.
//
//       URL:  www.brother.com
//


#include <linux/usb.h>

#include <linux/uaccess.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/slab.h>
//
#include <asm/ioctl.h>


#include "brfax_service.h"
#define _BRFAX_


#if 1
#define TCGETS		0x5401
#define TCSETS		0x5402
#define TCSETSW		0x5403
#define TCSETSF		0x5404
#define TCGETA		0x5405
#define TCSETA		0x5406
#define TCSETAW		0x5407
#define TCSETAF		0x5408
#define TCSBRK		0x5409
#define TCXONC		0x540A
#define TCFLSH		0x540B
#define TIOCEXCL	0x540C


typedef unsigned char	cc_t;
typedef unsigned int	speed_t;
typedef unsigned int	tcflag_t;

#define NCCS 32
#define VTIME 5
#define VMIN 6
struct termios
  {
    tcflag_t c_iflag;
    tcflag_t c_oflag;
    tcflag_t c_cflag;
    tcflag_t c_lflag;
    cc_t c_line;
    cc_t c_cc[NCCS];
    speed_t c_ispeed;
    speed_t c_ospeed;
#define _HAVE_STRUCT_TERMIOS_C_ISPEED 1
#define _HAVE_STRUCT_TERMIOS_C_OSPEED 1
  };

#define NCC 8
struct termio {
	unsigned short c_iflag;
	unsigned short c_oflag;
	unsigned short c_cflag;
	unsigned short c_lflag;
	unsigned char c_line;
	unsigned char c_cc[NCC];
};
#endif


MODULE_AUTHOR( "Brother" );
MODULE_DESCRIPTION( "Brother MFC modem class driver." );
MODULE_LICENSE( "GPL" );


#define      CLIENT_IS_APPR    0
#define      CLIENT_IS_DEVICE  1

int   device_major_num = 0;
static char* device_name  = "brusbmfc";
static char* msg      = "module [brusbmfc.o]";

#define BRTCOOFF 0
#define BRTCOON  1
#define BRTCIOFF 2
#define BRTCION  3
static int tco  = BRTCOON;
static int tci  = BRTCION;

static char vtime;
static char vmin;



char br_termios_init[] ={
  0x00,0x05,0x00,0x00,0x05,0x00,0x00,0x00,
  0xbd,0x0c,0x00,0x00,0x3b,0x8a,0x00,0x00,
  0x00,0x03,0x1c,0x7f,0x15,0x04,0x00,0x01,
  0x00,0x11,0x13,0x1a,0x00,0x12,0x0f,0x17,
  0x16,0x00,0x00,0x00,                    
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

#define br_termios_len  36

char br_termio_init[] ={
  0x00,0x05,0x05,0x00,0xbd,0x0c,0x3b,0x8a,
  0x00,0x03,0x1c,0x7f,0x15,0x04,0x00,0x01,
  0x00,                                   
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

#define br_termio_len  17


char br_termios[sizeof(struct termios)];
char br_termio [sizeof(struct termio)];


#define MAXBUF 1024

#define APPR  0
#define DEV   1


#define MAX_PIPE 2
#define PIPE_DEV2APP  0
#define PIPE_APP2DEV  1


static unsigned char *devbuf[MAX_PIPE];
static int buf_pos[MAX_PIPE];
static struct semaphore brusbmfc_sem;
static wait_queue_head_t write_queue[MAX_PIPE];
static wait_queue_head_t read_queue[MAX_PIPE];

static DECLARE_WAIT_QUEUE_HEAD(wq);

static int force_exit;

static int app_visited = 0;
static spinlock_t my_lock = SPIN_LOCK_UNLOCKED;

#define KL_ISCLIENTAPPR 0
#define KL_RELEASE      1
#define KL_OPEN         1
#define KL_LOAD         1
#define KL_UNLOAD       1
#define KL_READ         0
#define KL_WRITE        0
#define KL_POLL         0
#define KL_IOCTL        0


static void br_kernellog(int flag, char* statement )
{
	if(flag == 0)
		return;

	if ( current != NULL )
	{
		printk( KERN_INFO "%s(%-5i) :cmdline=\"%-10s\" \n          : %s\n", 
			msg, current->pid, current->comm, statement );
	}
	else
	{
		printk( KERN_INFO "%s : %s\n", msg, statement );
	}
}
#define  CONDITION_D_AVAIL_A2D()   (buf_pos[PIPE_APP2DEV]!=0)
#define  CONDITION_D_AVAIL_D2A()   (buf_pos[PIPE_DEV2APP]>=vmin)


#define  CONDITION_D_FULL_A2D()    (buf_pos[PIPE_APP2DEV]>=(MAXBUF - 64))
#define  CONDITION_D_FULL_D2A()    (buf_pos[PIPE_DEV2APP]==MAXBUF)



#define CONDITION_APP_READ()  CONDITION_D_AVAIL_D2A()
#define CONDITION_DEV_READ()  CONDITION_D_AVAIL_A2D()

#define CONDITION_APP_WRITE() (!CONDITION_D_FULL_A2D())
#define CONDITION_DEV_WRITE() (!CONDITION_D_FULL_D2A())




#define CONDITION_POLL_APP_READ() CONDITION_D_AVAIL_D2A()
#define CONDITION_POLL_DEV_READ() CONDITION_D_AVAIL_A2D()

#define CONDITION_POLL_APP_WRITE() (!CONDITION_D_FULL_A2D())
#define CONDITION_POLL_DEV_WRITE() (!CONDITION_D_FULL_D2A())




#define CONDITION_READ()    ((force_exit==0)&&(tci==BRTCION)&&((client == CLIENT_IS_DEVICE)? \
					       CONDITION_DEV_READ():CONDITION_APP_READ()))
#define CONDITION_WRITE()   ((force_exit==0)&&((client == CLIENT_IS_DEVICE)? \
					       CONDITION_DEV_WRITE():CONDITION_APP_WRITE()))





#define CONDITION_POLL_READ()    ((force_exit==0)&&(tci==BRTCION)&&((client == CLIENT_IS_DEVICE)? \
						    CONDITION_POLL_DEV_READ():CONDITION_POLL_APP_READ()))
#define CONDITION_POLL_WRITE()   ((force_exit==0)&&((client == CLIENT_IS_DEVICE)? \
						    CONDITION_POLL_DEV_WRITE():CONDITION_POLL_APP_WRITE()))

int is_client_apprication(void){
  if( strstr(current->comm , BRFAX_DAEMON_NAME) ){
    return 0;
  } 
  else{ 
    return 1;
  }
  return 0;
}


static int 
brusbmfc_open( struct inode* inode, struct file* p_file )
{
#ifdef	_BRFAX_
	char msg_buf[256];
	int msg_len = 0;
#endif
	if( ! is_client_apprication())
	{
		force_exit = 0;
		buf_pos[PIPE_APP2DEV] = 0;
		buf_pos[PIPE_DEV2APP] = 0;
        spin_lock(&my_lock);
        app_visited = 0;
        spin_unlock(&my_lock);
	}
#ifdef	_BRFAX_
	else 
	{
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);
		spin_lock(&my_lock);		
		if( app_visited != 0)
		{
			spin_unlock(&my_lock);
			return -EBUSY;
		}
		app_visited = 1;
		spin_unlock(&my_lock);
		
		msg_len = sprintf(msg_buf, BRMSG_APP_START_FORMAT, current->pid);
		if ( msg_len > ( MAXBUF - buf_pos[PIPE_APP2DEV] ))
		{
			spin_lock(&my_lock);
			app_visited = 0;
			spin_unlock(&my_lock);
			return -EFAULT;
		}

		memcpy( devbuf[PIPE_APP2DEV] + buf_pos[PIPE_APP2DEV], msg_buf, msg_len );
		buf_pos[PIPE_APP2DEV] += msg_len;

		
	}
#endif	// _BRFAX_
	br_kernellog( KL_OPEN , "brusbmfc_open()" );

	return 0;
}


static int 
brusbmfc_release( struct inode* inode, struct file* p_file )
{
#ifdef	_BRFAX_
	char msg_buf[256];
	int msg_len = 0;
#endif
  if( ! is_client_apprication()){
    char ss[256];
    sprintf(ss,
	    "\nbrusbmfc_close()\n"
	    "buf_pos[PIPE_APP2DEV] = %d\n"
	    "buf_pos[PIPE_DEV2APP] = %d\n"
	    "force_exit = %d\n",
	    buf_pos[PIPE_APP2DEV],
	    buf_pos[PIPE_DEV2APP],
	    force_exit);
    br_kernellog( KL_RELEASE,ss );
    force_exit = 1;
    tci = BRTCION;
    tco = BRTCOON;
    wake_up_interruptible( &write_queue[PIPE_APP2DEV] );
    wake_up_interruptible( &write_queue[PIPE_DEV2APP] );
  }
  else{
    tci = BRTCION;
#ifdef	_BRFAX_
//	if( app_visited != 0 )
//	{
		msg_len = sprintf(msg_buf, BRMSG_APP_EXIT);
		if ( msg_len > ( MAXBUF - buf_pos[PIPE_APP2DEV] ))
		{
			spin_lock(&my_lock);
			app_visited = 0;
			spin_unlock(&my_lock);
			return -EFAULT;
		}
		memcpy( devbuf[PIPE_APP2DEV] + buf_pos[PIPE_APP2DEV], msg_buf, msg_len );
		buf_pos[PIPE_APP2DEV] += msg_len;
//		app_visited = 0;
//	}
#endif	// _BRFAX_
    br_kernellog( KL_RELEASE,"brusbmfc_close()" );
  }
  return 0;
}


static ssize_t 
brusbmfc_read( struct file* p_file, char* buf, size_t count, loff_t* pos ){
  int copy_len;
  int i;
  int selected_pipe;
  int client;
  long timeout;
  long rcwe ;
  br_kernellog( KL_READ, "brusbmfc_read()" );

  
  if( force_exit == 1 ){
    br_kernellog( KL_READ, "brusbmfc_read() : brfaxd is not started" );
    return -ERESTARTSYS;
  }

  if(is_client_apprication()){
    selected_pipe  = PIPE_DEV2APP;
    client = CLIENT_IS_APPR;
  }
  else{
    selected_pipe  = PIPE_APP2DEV;
    client = CLIENT_IS_DEVICE;
  }


  if ( down_interruptible( &brusbmfc_sem ) ) {
    br_kernellog(KL_READ,  "brusbmfc_read:down_interruptible failed (1)" );
    return -ERESTARTSYS;
  }
  
  
  timeout = 0;
  while ( (!CONDITION_READ() ) && (timeout == 0)) {
    up( &brusbmfc_sem );
    
    if ( p_file->f_flags & O_NONBLOCK ) {
      br_kernellog(KL_READ,  "brusbmfc_read:read return (O_NONBLOCK)" );
      return -EAGAIN;
    }
    
#if KL_READ == 2
    {
      char ss[256];
      sprintf(ss,
	      "\nbrusbmfc_read(): goto sleep\n"
	      "buf_pos[PIPE_APP2DEV] = %d\n"
	      "buf_pos[PIPE_DEV2APP] = %d\n"
	      "force_exit = %d\n",
	      buf_pos[PIPE_APP2DEV],
	      buf_pos[PIPE_DEV2APP],
	      force_exit);

      br_kernellog( KL_READ,ss );
    }
#else  //KL_READ == 2
    br_kernellog( KL_READ,"brusbmfc_read:sleep" );
#endif // KL_READ == 2 

    if(client != CLIENT_IS_APPR){
      if ( wait_event_interruptible( read_queue[selected_pipe], 
				     CONDITION_READ()
				     ) ){
	return -ERESTARTSYS;
      }
    }else{
      rcwe = wait_event_interruptible_timeout( read_queue[selected_pipe], 
					       CONDITION_READ() , 
					       (vtime==0&&vmin>0)?-1:vmin );
      if(rcwe == -ERESTARTSYS){
	return -ERESTARTSYS;
      }
      else if(rcwe == 0){
	br_kernellog( KL_READ, "brusbmfc_read: TIMEOUT" );
	timeout = 1;
      }
    }

    if ( down_interruptible( &brusbmfc_sem ) ) {
      br_kernellog( KL_READ, "brusbmfc_read:down_interruptible failed (2)" );
      return -ERESTARTSYS;
    }
  }    
  
  //____________________________________________________________
  if ( count > buf_pos[selected_pipe] )
    copy_len = buf_pos[selected_pipe];
  else
    copy_len = count;

#if KL_READ == 2
  {
    char ss[256];
    sprintf(ss, "brusbmfc_read:read start (%d)",copy_len );
    br_kernellog( KL_READ,ss );
  }
#else //KL_READ == 2
  br_kernellog( KL_READ,"brusbmfc_read:read start" );
#endif //KL_READ == 2
  
  if ( copy_to_user( buf, devbuf[selected_pipe], copy_len ) ) {
    up( &brusbmfc_sem );
    br_kernellog( KL_READ,"copy_to_user failed" );
    return -EFAULT;
  }
  
  *pos += copy_len;
  
  for ( i = copy_len; i < buf_pos[selected_pipe]; i ++ )
    devbuf[selected_pipe][ i - copy_len ] = devbuf[selected_pipe][i];
  
  buf_pos[selected_pipe] -= copy_len;
  br_kernellog( KL_READ, "brusbmfc_read:copy end" );

  //____________________________________________________________
  
  
  up( &brusbmfc_sem );
  wake_up_interruptible( &write_queue[selected_pipe] );
#if KL_READ == 2
  {
    char ss[256];
    sprintf(ss,
	    "\nbrusbmfc_read(): wake_up_interruptible( &write_queue[%d] );\n"
	    "buf_pos[PIPE_APP2DEV] = %d\n"
	    "buf_pos[PIPE_DEV2APP] = %d\n"
	    "force_exit = %d\n"
	    "copy_len/count = %d/%d",
	    selected_pipe,
	    buf_pos[PIPE_APP2DEV],
	    buf_pos[PIPE_DEV2APP],
	    force_exit,
	    copy_len,count);
    br_kernellog( KL_READ,ss );
  }
#endif //KL_READ == 2
  
  if( client == CLIENT_IS_DEVICE )	//brfaxd
  {
	  if( strstr(buf, BRMSG_APP_EXIT) )
	  {
		  spin_lock(&my_lock);
		  app_visited = 0;
		  spin_unlock(&my_lock);
	  }

  }

  br_kernellog( KL_READ,"brusbmfc_read:function end" );

  return copy_len;
}


static ssize_t 
brusbmfc_write( struct file* p_file, const char* buf, size_t count, loff_t* pos ){
  int copy_len;
  int selected_pipe;
  int client;

#if KL_WRITE == 1
  br_kernellog( KL_WRITE,"brusbmfc_write()" );
#endif //KL_WRITE == 1


  if( force_exit == 1 ){
    br_kernellog( KL_READ, "brusbmfc_write() : brfaxd is not started" );
    return -ERESTARTSYS;
  }

  if(is_client_apprication()){
    selected_pipe  = PIPE_APP2DEV;
    client = CLIENT_IS_APPR;
  }
  else{
    selected_pipe  = PIPE_DEV2APP;
    client = CLIENT_IS_DEVICE;
  }

#if KL_WRITE == 2
  {
    char ss[256];
    sprintf(ss,"brusbmfc_write() from %s  size = %d remain = %d",
	    (client = CLIENT_IS_APPR)?"APPR":"DEVICE",
	    count,buf_pos[selected_pipe] );
    br_kernellog( KL_WRITE,ss );
  }
#endif //KL_WRITE == 2

  if ( down_interruptible( &brusbmfc_sem ) ) {
    br_kernellog( KL_WRITE,"down_interruptible for write failed (1)" );
    return -ERESTARTSYS;
  }

  
  while ( !CONDITION_WRITE() ){

    up( &brusbmfc_sem );
    
    if ( p_file->f_flags & O_NONBLOCK ) {
      br_kernellog( KL_WRITE,"brusbmfc_write:return (O_NONBLOCK)" );
      return -EAGAIN;
    }
#if KL_WRITE == 2
    {
      char ss[256];
      sprintf(ss,
	      "\nbrusbmfc_write(): goto sleep\n"
	      "buf_pos[PIPE_APP2DEV] = %d\n"
	      "buf_pos[PIPE_DEV2APP] = %d\n"
	      "force_exit = %d\n",
	      buf_pos[PIPE_APP2DEV],
	      buf_pos[PIPE_DEV2APP],
	      force_exit);
      br_kernellog( KL_RELEASE,ss );
    }
#else // KL_WRITE == 2
    br_kernellog( KL_WRITE,"brusbmfc_write:sleep" );
#endif // KL_WRITE == 2
    
    if ( wait_event_interruptible( write_queue[selected_pipe], CONDITION_WRITE() )){
      br_kernellog( KL_WRITE,"brusbmfc_write:wait_event_interruptible failed (2)" );
      return -ERESTARTSYS;
    }

    br_kernellog( KL_WRITE,"brusbmfc_write:wakeup" );
    
    if ( down_interruptible( &brusbmfc_sem ) ) {
      br_kernellog( KL_WRITE,"brusbmfc_write:down_interruptible failed (3)" );
      return -ERESTARTSYS;
    }
  }
  

  br_kernellog( KL_WRITE,"brusbmfc_write:write start" );



  if( force_exit == 0 ){
    //___________________________________________________________
    if ( count > ( MAXBUF - buf_pos[selected_pipe] ) )
      copy_len = MAXBUF - buf_pos[selected_pipe];
    else
      copy_len = count;

    if ( copy_from_user( devbuf[selected_pipe] + buf_pos[selected_pipe], 
			 buf, copy_len ) ) {
      up( &brusbmfc_sem );
      br_kernellog( KL_WRITE,"copy_from_user failed" );
      return -EFAULT;
    }
    *pos    += copy_len;
    buf_pos[selected_pipe] += copy_len;
    br_kernellog( KL_WRITE,"brusbmfc_write: copy end" );
    //___________________________________________________________
  }
  else{
    copy_len = count;
    *pos    += copy_len;
  }

  up( &brusbmfc_sem );
  wake_up_interruptible( &read_queue[selected_pipe] );

  br_kernellog( KL_WRITE,"brusbmfc_write: end" );
  return copy_len;
}


static unsigned int 
brusbmfc_poll( struct file* p_file, poll_table* wait )
{
  unsigned int retflag = 0;
  int client;

  if(is_client_apprication()){
    client = CLIENT_IS_APPR;
    poll_wait( p_file, &read_queue[PIPE_DEV2APP],  wait );
    poll_wait( p_file, &write_queue[PIPE_APP2DEV], wait );
  }
  else{
    client = CLIENT_IS_DEVICE;
    poll_wait( p_file, &read_queue[PIPE_APP2DEV],  wait );
    poll_wait( p_file, &write_queue[PIPE_DEV2APP], wait );
  }

  if(CONDITION_POLL_READ()){
    retflag |= ( POLLIN  | POLLRDNORM );
  }
  if(CONDITION_POLL_WRITE()){
    retflag |= ( POLLOUT | POLLWRNORM );
  }

#if KL_POLL == 1
  if (retflag & ( POLLOUT | POLLWRNORM |POLLIN  | POLLRDNORM )){
    char ss[256], *cl, *rstate, *wstate;
    cl = (client == CLIENT_IS_APPR)? "APR":"DEV";
    rstate = (retflag &( POLLIN  | POLLRDNORM ))? "Allow" : "Denied";
    wstate = (retflag &( POLLOUT | POLLWRNORM ))? "Allow" : "Denied";
    
    sprintf(ss, "brusbmfc_poll : from %s read = %s write = %s",
	    cl , rstate , wstate);
    br_kernellog( KL_POLL,ss );
  }
#endif // KL_POLL == 1
#if KL_POLL == 2
  {
    char ss[256], *cl, *rstate, *wstate;
    cl = (client == CLIENT_IS_APPR)? "APR":"DEV";
    rstate = (retflag &( POLLIN  | POLLRDNORM ))? "Allow" : "Denied";
    wstate = (retflag &( POLLOUT | POLLWRNORM ))? "Allow" : "Denied";
    
    sprintf(ss, "brusbmfc_poll : from %s   read = %s write = %s",
	    cl , rstate , wstate);
    br_kernellog( KL_POLL,ss );
  }

#endif // KL_POLL == 2
  return retflag;
}

#ifdef	HAVE_UNLOCKED_IOCTL
static long brusbmfc_ioctl( 
//	struct inode* inode, 
	struct file* p_file, 
	unsigned int cmd, 
	unsigned long arg )
#else
static int brusbmfc_ioctl( 
	struct inode* inode, 
	struct file* p_file, 
	unsigned int cmd, 
	unsigned long arg )
#endif
{

  br_kernellog( KL_IOCTL, "brusbmfc_ioctl()" );

  switch( cmd ) {

  case TCGETS:
    br_kernellog(KL_IOCTL, "brusbmfc_ioctl: control=TCGETS" );
    copy_to_user((char *)arg , br_termios ,br_termios_len ); 
    vmin =   ((struct termios *)br_termios)->c_cc[VMIN] ;
    vtime =  ((struct termios *)br_termios)->c_cc[VTIME];

    ((struct termio *)br_termio)->c_cc[VMIN] = vmin;
    ((struct termio *)br_termio)->c_cc[VTIME] = vtime;

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCGETS" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCGETS vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif    
    br_kernellog(KL_IOCTL,"copy end");
    break;

  case TCSETS:
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETS" );
    copy_from_user(br_termios , (struct termios *)arg, sizeof(br_termios)); 
    vmin =   ((struct termios *)br_termios)->c_cc[VMIN] ;
    vtime =  ((struct termios *)br_termios)->c_cc[VTIME];
    ((struct termio *)br_termio)->c_cc[VMIN] = vmin;
    ((struct termio *)br_termio)->c_cc[VTIME] = vtime;

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETS" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCSETS vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif    

    break;
  case TCSETSW:
    copy_from_user(br_termios , (struct termios *)arg, sizeof(struct termios)); 
    vmin =   ((struct termios *)br_termios)->c_cc[VMIN] ;
    vtime =  ((struct termios *)br_termios)->c_cc[VTIME];
    ((struct termio *)br_termio)->c_cc[VMIN] = vmin;
    ((struct termio *)br_termio)->c_cc[VTIME] = vtime;

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETSW" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCSETSW vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif    

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETSW" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCSETSW vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif
    break;
  case TCSETSF:
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETSF" );
    copy_from_user(br_termios , (struct termios *)arg, sizeof(struct termios)); 
    vmin =   ((struct termios *)br_termios)->c_cc[VMIN] ;
    vtime =  ((struct termios *)br_termios)->c_cc[VTIME];
    ((struct termio *)br_termio)->c_cc[VMIN] = vmin;
    ((struct termio *)br_termio)->c_cc[VTIME] = vtime;

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETSF" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCSETSF vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif    
    break;


    //---------------

  case TCGETA:
    copy_to_user((char *)arg ,br_termio, br_termio_len);
    vmin =   ((struct termio *)br_termio)->c_cc[VMIN] ;
    vtime =  ((struct termio *)br_termio)->c_cc[VTIME];
    ((struct termios *)br_termios)->c_cc[VMIN] = vmin;
    ((struct termios *)br_termios)->c_cc[VTIME] = vtime;

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCGETA" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCGETA vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif
    break;
  case TCSETA:
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETA" );
    copy_from_user(br_termio , (struct termio *)arg, sizeof(struct termio)); 
    vmin =   ((struct termio *)br_termio)->c_cc[VMIN] ;
    vtime =  ((struct termio *)br_termio)->c_cc[VTIME];
    ((struct termios *)br_termios)->c_cc[VMIN] = vmin;
    ((struct termios *)br_termios)->c_cc[VTIME] = vtime;

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETA" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCSET vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif

    break;



  case TCSETAW:
    copy_from_user(br_termio , (struct termio *)arg, sizeof(struct termio)); 
    vmin =   ((struct termio *)br_termio)->c_cc[VMIN] ;
    vtime =  ((struct termio *)br_termio)->c_cc[VTIME];
    ((struct termios *)br_termios)->c_cc[VMIN] = vmin;
    ((struct termios *)br_termios)->c_cc[VTIME] = vtime;

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETAW" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCSETW vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif
    break;
  case TCSETAF:
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETAF" );
    copy_from_user(br_termio , (struct termio *)arg, sizeof(struct termio)); 
    vmin =   ((struct termio *)br_termio)->c_cc[VMIN] ;
    vtime =  ((struct termio *)br_termio)->c_cc[VTIME];
    ((struct termios *)br_termios)->c_cc[VMIN] = vmin;
    ((struct termios *)br_termios)->c_cc[VTIME] = vtime;

#if KL_IOCTL==1
    br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCSETAF" );
#elif KL_IOCTL==2
    {
      char ss[256];
      sprintf(ss, 
	      "brusbmfc_ioctl: control=TCSETF vmin=%d vtime=%d\n" ,
	      vmin,vtime);
      br_kernellog( KL_IOCTL, ss );
    }
#endif
    break;

  case TCXONC:	
    if( is_client_apprication()){
      switch(arg){
      case BRTCOOFF:
	br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCXONC arg=BRTCOOFF" );
	tco  = arg;
	break;
      case BRTCOON:
	br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCXONC arg=BRTCOON" );
	tco  = arg;
	break;
      case BRTCIOFF:
	br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCXONC arg=BRTCIOFF" );
	tci  = arg;
	break;
      case BRTCION:
	br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCXONC arg=BRTCION" );
	tci  = arg;
	break;
      default:
	br_kernellog( KL_IOCTL,"brusbmfc_ioctl: control=TCXONC arg=other" );
	break;
      }
    }
    break;

    
  default:
    {
      char   message[64];
      sprintf(message ,"brusbmfc_ioctl: UNKNOWN control=%x" , cmd);
      br_kernellog( KL_IOCTL,message );
    }
    
    return -EINVAL;
  }
  
  br_kernellog( KL_IOCTL,"brusbmfc_ioctl(): end" );

  return 0;
}


static struct file_operations brusbmfc_fops = 
{
 owner   : THIS_MODULE,
 read    : brusbmfc_read,
 write   : brusbmfc_write,
 poll    : brusbmfc_poll,
#ifdef	HAVE_UNLOCKED_IOCTL
 unlocked_ioctl   : brusbmfc_ioctl,
#else
 ioctl   : brusbmfc_ioctl,
#endif
 open    : brusbmfc_open,
 release : brusbmfc_release,
};


int init_module( void ){
  int i;
  device_major_num = register_chrdev( 
				     0, 
				     device_name, 
				     &brusbmfc_fops );
  if(device_major_num == 0){
    br_kernellog( KL_LOAD,"register_chrdev failed" );
    return -EBUSY;
  }
  force_exit = 1;
  devbuf[PIPE_DEV2APP] = kmalloc(MAXBUF,GFP_KERNEL);
  devbuf[PIPE_APP2DEV] = kmalloc(MAXBUF,GFP_KERNEL);

  buf_pos[PIPE_DEV2APP] = 0;
  buf_pos[PIPE_APP2DEV] = 0;

  if(devbuf[PIPE_DEV2APP] == 0 || devbuf[PIPE_APP2DEV] == 0){
    if(devbuf[PIPE_DEV2APP])kfree(devbuf[PIPE_DEV2APP]);
    if(devbuf[PIPE_APP2DEV])kfree(devbuf[PIPE_APP2DEV]);
    br_kernellog( KL_LOAD, "memory allocation failed" );
    return -ENOMEM;
  }

  sema_init( &brusbmfc_sem, 1 );

  init_waitqueue_head( &write_queue[PIPE_DEV2APP] );
  init_waitqueue_head( &read_queue[PIPE_DEV2APP] );
  init_waitqueue_head( &write_queue[PIPE_APP2DEV] );
  init_waitqueue_head( &read_queue[PIPE_APP2DEV] );




  for(i=0; i<br_termio_len         ;  i ++)br_termio[i] = br_termio_init[i];
  for(   ; i<sizeof(struct termio) ;  i ++)br_termio[i] = 0;

  for(i=0; i<br_termios_len        ;  i ++)br_termios[i] = br_termios_init[i];
  for(   ; i<sizeof(struct termios);  i ++)br_termios[i] = 0;

  vmin =   ((struct termios *)br_termios)->c_cc[VMIN] ;
  vtime =  ((struct termios *)br_termios)->c_cc[VTIME];

  br_kernellog( KL_LOAD,"loaded  into kernel" );
  return 0;
}


void
cleanup_module( void )
{

  force_exit = 1;

  if(devbuf[PIPE_DEV2APP])kfree(devbuf[PIPE_DEV2APP]);
  if(devbuf[PIPE_APP2DEV])kfree(devbuf[PIPE_APP2DEV]);

  unregister_chrdev( device_major_num, device_name );

}


/* End of brusbmfc.c */
